// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"crypto/x509"
	"encoding/pem"
	"io/ioutil"
	"os"

	"github.com/spf13/cobra"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

// registerCmd represents the "register" command.
var registerCmd = &cobra.Command{
	Use:   "register",
	Short: "Generate a CSR and ask Xbus for a registration.",
	Long: `Generate a Certificate Signing Request then send it over to Xbus through the
anonymouse "registration" channel.`,
	Run: register,
}

// register parses arguments to proceed with the registration.
func register(cmd *cobra.Command, args []string) {
	os.Exit(doRegister())
}

// loadPrivKey loads the private key stored in the specified file.
func loadPrivKey(privKeyPath string) (privKey interface{}, err error) {
	privKeyData, err := ioutil.ReadFile(privKeyPath)
	if err != nil {
		return nil, err
	}
	privKeyBlock, _ := pem.Decode(privKeyData)
	return x509.ParseECPrivateKey(privKeyBlock.Bytes)
}

// doRegister generates a CSR then asks Xbus for a registration.
func doRegister() int {
	xbusClient, err := GetClient(true)
	if err != nil {
		Logger.Errorf("%s", err)
		return 1
	}

	if err := xbusClient.CreateKey(false); err != nil {
		if err != xbus.ErrAlreadyExists {
			Logger.Errorf("Error creating the private key: %s", err)
			return 1
		}
	}
	// TODO find the actual hostname or get one in the configuration
	// TODO Read org and cn from the configuration
	if err := xbusClient.CreateCSR("localhost", "xbus", "xbusctl", false); err != nil {
		if err != xbus.ErrAlreadyExists {
			Logger.Errorf("Cannot generate a CSR: %s", err)
			return 1
		}
	}

	// Send the registration request.
	regStatus, err := xbusClient.Register()
	if err != nil {
		Logger.Errorf("Cannot send the registration request: %s", err)
		return 1
	}

	Logger.Noticef("Successfully sent the registration request.")

	// Say something depending on the registration status we have received in reply.
	switch regStatus {
	case api.Registration_PENDING:
		Logger.Noticef(`The registration is pending approval. To approve from Xbus, run:
	xbusd account accept %q`,
			//tools.GetCSRHash(string(csrData), false),
			"xxxx",
		)
	case api.Registration_ACCEPTED:
		Logger.Noticef(
			"The registration was already known by Xbus; it has been automatically accepted.",
		)
	case api.Registration_REJECTED:
		Logger.Noticef(
			"The registration was already known by Xbus; it has been automatically rejected.",
		)
	}

	return 0
}

func init() {
	RootCmd.AddCommand(registerCmd)
}
