// Copyright 2016 Orus.io. All rights reserved.

package fullstack

import (
	"context"
	"fmt"
	"testing"
	"time"

	_ "github.com/mattn/go-sqlite3"

	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/api/control"
	"bitbucket.org/orus-io/xbus/testutils"
)

func TestStartupShutdown(t *testing.T) {
	ft := InitFullstack(t, true)
	defer ft.Shutdown()
}

func TestEmptyOutputRequest(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	ft.Logger.ExpectsError(testutils.ExpectsContains(
		"[server] [broker] CLIENT error: Missing envelope"))

	_, err := emitter.Client.API.NewBroker(
		emitter.ID.String(),
	).Output(
		api.OutputRequest{},
	)
	assert.NotNil(t, err)
	assert.Equal(t, "CLIENT error: Missing envelope", err.Error())
}

func TestEndToEndSingleEventBasics(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	xsub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			t.Logf("Here i am")
			r, err := apc.ReadEnvelope("default")
			if err != nil {
				ft.T.Error(err)
				return err
			}
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return err
			}
			hit <- e
			return nil
		})
	assert.Nil(t, err)
	defer xsub.Unsubscribe()

	assert.Equal(t, "jojo", consumer.Name)

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))

	env := xbus.ForgeEnvelopeFromItems("evt1", []api.Item{[]byte("\"item1\"")})

	envStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	status, err := emitter.SendEnvelope(nil, api.DefaultOutput, env)
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	timeout := time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("Process state not updated")
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_RUNNING, envState.Status)
	}

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("Process state not updated to 'done'")
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_DONE, envState.Status)
	}
}

func TestEndToEndSingleEventWithConsumerDown(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	var handler = func(apc *xbus.ActorProcessingContext) error {
		r, err := apc.ReadEnvelope("default")
		if err != nil {
			ft.T.Error(err)
			return err
		}
		e, err := r.Complete(time.Second)
		if err != nil {
			ft.T.Error(err)
			return err
		}
		hit <- e
		return nil
	}
	xsub, err := consumer.Subscribe(handler)
	if err != nil {
		t.Fatal(err)
	}
	defer func() {
		xsub.Unsubscribe()
	}()

	assert.Equal(t, "jojo", consumer.Name)

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))

	// take the consumer offline
	xsub.Unsubscribe()

	status, err := emitter.SendItems(nil, api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	// wait enough for the consumer to be declared 'off'
	// It should be pretty fast if the server handle 'ActorLeaving' in a decent time
	timeout := time.After(300 * time.Millisecond)
	select {
	case <-timeout:
		// everything is normal
		actor, exists := ft.Server.ActorStorage().Get(consumer.ID)
		assert.True(t, exists)
		assert.False(t, actor.Online)
		assert.False(t, actor.Unresponsive)
	case <-hit:
		t.Fatalf("Very wrong, we unsubscribed the consumer, we should not receive anything (yet)")
	}

	assert.Nil(t, ft.Client.ProcessQuery(
		context.Background(),
		api.ProcessFilter{Status: []api.Process_Status{api.Process_RUNNING}},
		func(pList []api.Process) {
			assert.Equal(t, 1, len(pList))
			assert.Equal(t, api.Process_RUNNING, pList[0].Status)
		},
	))

	// go back online
	xsub, err = consumer.Subscribe(handler)
	assert.Nil(t, err)

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}
}

func TestEndToEndSingleEventWithConsumerUnresponsive(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	var handler = func(apc *xbus.ActorProcessingContext) error {

		r, err := apc.ReadEnvelope("default")
		if err != nil {
			ft.T.Error(err)
			return err
		}
		e, err := r.Complete(time.Second)
		if err != nil {
			ft.T.Error(err)
			return err
		}
		hit <- e
		return nil
	}
	// First set a buggy consumer subscription (ie, none)
	consumer.Client.API.Sentinel.ActorReady(consumer.ID.String())

	ft.Logger.ExpectsError(testutils.ExpectsContains(
		"[server] [director] Agent jojo unresponsive (1 time)"))
	ft.Logger.ExpectsError(testutils.ExpectsContains(
		"[server] [director] Agent jojo unresponsive (2 time)"))

	assert.Equal(t, "jojo", consumer.Name)

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))

	status, err := emitter.SendItems(nil, api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	// wait enough for the consumer to be declared 'unresponsive'
	select {
	case <-time.After(5500 * time.Millisecond):
		// everything is normal
		actor, exists := ft.Server.ActorStorage().Get(consumer.ID)
		assert.True(t, exists)
		assert.True(t, actor.Online)
		assert.True(t, actor.Unresponsive)
	case <-hit:
		t.Fatalf("Very wrong, we unsubscribed the consumer, we should not receive anything (yet)")
	}

	assert.Nil(t, ft.Client.ProcessQuery(
		context.Background(),
		api.ProcessFilter{Status: []api.Process_Status{api.Process_RUNNING}},
		func(pList []api.Process) {
			assert.Equal(t, 1, len(pList))
			assert.Equal(t, api.Process_RUNNING, pList[0].Status)
		},
	))

	// now setup a serious subscriber, and let it handle the next retry
	xsub, err := consumer.Subscribe(handler)
	if err != nil {
		t.Fatal(err)
	}
	defer xsub.Unsubscribe()

	select {
	case <-time.After(30 * time.Second):
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
		actor, exists := ft.Server.ActorStorage().Get(consumer.ID)
		assert.True(t, exists)
		assert.True(t, actor.Online)
		assert.False(t, actor.Unresponsive)
	}
}

func TestEndToEndSingleEventWithConsumerNeverResponsive(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})

	// Set a buggy consumer subscription (ie, none)
	consumer.Client.API.Sentinel.ActorReady(consumer.ID.String())

	ft.Logger.ExpectsError(testutils.ExpectsContains(
		"[server] [director] Agent jojo unresponsive (1 time)"))
	ft.Logger.ExpectsError(testutils.ExpectsContains(
		"[server] [director] Agent jojo unresponsive (2 time)"))
	ft.Logger.ExpectsError(testutils.ExpectsContains(
		"[server] [director] Agent jojo unresponsive (3 time)"))
	ft.Logger.ExpectsError(testutils.ExpectsContains(
		"[server] [director] Agent jojo unresponsive (4 time)"))
	ft.Logger.ExpectsError(testutils.ExpectsContains(
		"[server] [director] Agent jojo unresponsive (5 time)"))

	assert.Equal(t, "jojo", consumer.Name)

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))

	status, err := emitter.SendItems(nil, api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	// wait enough for the consumer to be declared 'unresponsive', which is 50 ms in tests
	timeout := time.After(5500 * time.Millisecond)
	select {
	case <-timeout:
		// everything is normal
		actor, exists := ft.Server.ActorStorage().Get(consumer.ID)
		assert.True(t, exists)
		assert.True(t, actor.Online)
		assert.True(t, actor.Unresponsive)
	}

	assert.Nil(t, ft.Client.ProcessQuery(
		context.Background(),
		api.ProcessFilter{Status: []api.Process_Status{api.Process_RUNNING}},
		func(pList []api.Process) {
			assert.Equal(t, 1, len(pList))
			assert.Equal(t, api.Process_RUNNING, pList[0].Status)
		},
	))

	// Now wait enough for the actor to be marked offline
	timeout = time.After(30 * time.Second)
	select {
	case <-timeout:
		actor, exists := ft.Server.ActorStorage().Get(consumer.ID)
		assert.True(t, exists)
		assert.False(t, actor.Online)
		assert.True(t, actor.Unresponsive)
	}
}

func TestEndToEndSingleEventWithServerShutdown(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")
	xbusClient, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	handler := func(apc *xbus.ActorProcessingContext) error {
		r, err := apc.ReadEnvelope("default")
		if err != nil {
			ft.T.Error(err)
			return err
		}
		e, err := r.Complete(time.Second)
		if err != nil {
			ft.T.Error(err)
			return err
		}
		hit <- e
		return nil
	}

	assert.Equal(t, "jojo", consumer.Name)

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))

	xbusClient.Close()

	status, err := emitter.SendItems(nil, api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	// wait enough for the consumer to be declared 'off'
	timeout := time.After(300 * time.Millisecond)
	select {
	case <-timeout:
		// everything is normal
		actor, exists := ft.Server.ActorStorage().Get(consumer.ID)
		assert.True(t, exists)
		assert.False(t, actor.Online)
		assert.False(t, actor.Unresponsive)
	case <-hit:
		t.Fatalf("Very wrong, we unsubscribed the consumer, we should not receive anything (yet)")
	}

	// stop/start the server
	ft.ServerStop()
	assert.True(t, ft.ServerStart())

	// go back online
	assert.Nil(t, xbusClient.Connect())
	xsub, err := consumer.Subscribe(handler)
	if err != nil {
		t.Fatal(err)
	}
	defer xsub.Unsubscribe()

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}
}

func TestEndToEndSingleEventWithPauseResume(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	handler := func(apc *xbus.ActorProcessingContext) error {
		r, err := apc.ReadEnvelope("default")
		if err != nil {
			ft.T.Error(err)
			return err
		}
		e, err := r.Complete(time.Second)
		if err != nil {
			ft.T.Error(err)
			return err
		}
		hit <- e
		return nil
	}
	xsub, err := consumer.Subscribe(handler)
	if err != nil {
		t.Fatal(err)
	}
	defer func() {
		xsub.Unsubscribe()
	}()

	assert.Equal(t, "jojo", consumer.Name)

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))

	// take the consumer offline
	xsub.Unsubscribe()

	status, err := emitter.SendItems(nil, api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	// wait enough for the consumer to be declared 'off'
	timeout := time.After(300 * time.Millisecond)
	select {
	case <-timeout:
		// everything is normal
		actor, exists := ft.Server.ActorStorage().Get(consumer.ID)
		assert.True(t, exists)
		assert.False(t, actor.Online)
		assert.False(t, actor.Unresponsive)
	case <-hit:
		t.Fatalf("Very wrong, we unsubscribed the consumer, we should not receive anything (yet)")
	}

	var pList []api.Process
	assert.Nil(t, ft.Client.ProcessQuery(
		context.Background(),
		api.ProcessFilter{Status: []api.Process_Status{api.Process_RUNNING}},
		func(l []api.Process) {
			pList = append(pList, l...)
		},
	))
	assert.Equal(t, 1, len(pList))
	assert.Equal(t, api.Process_RUNNING, pList[0].Status)

	err = ft.Client.ProcessControl(pList[0].ID, control.ProcessControlRequest_PAUSE)
	assert.Nil(t, err)

	// go back online
	xsub, err = consumer.Subscribe(handler)
	assert.Nil(t, err)

	timeout = time.After(500 * time.Millisecond)
	select {
	case <-timeout:
		// everything is normal
		actor, exists := ft.Server.ActorStorage().Get(consumer.ID)
		assert.True(t, exists)
		assert.True(t, actor.Online)
		assert.False(t, actor.Unresponsive)
	case <-hit:
		t.Fatalf("Very wrong, we paused the process, we should not receive anything")
	}

	err = ft.Client.ProcessControl(pList[0].ID, control.ProcessControlRequest_RESUME)
	assert.Nil(t, err)

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}
}

func TestEndToEndSingleEventWithConsumerError(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	xsub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			r, err := apc.ReadEnvelope("default")
			if err != nil {
				ft.T.Error(err)
				return err
			}
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return err
			}
			hit <- e
			return fmt.Errorf("bad thing happened: %s", "I just want not to work")
		})
	if err != nil {
		t.Fatal(err)
	}
	defer xsub.Unsubscribe()

	assert.Equal(t, "jojo", consumer.Name)

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))

	env := xbus.ForgeEnvelopeFromItems("evt1", []api.Item{[]byte("\"item1\"")})

	envStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	status, err := emitter.SendEnvelope(nil, api.DefaultOutput, env)
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	timeout := time.After(5 * time.Second)
	var processID api.UUID
	select {
	case <-timeout:
		t.Errorf("Process not running")
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_RUNNING, envState.Status)

		processID = envState.ProcessID
	}

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("Process not ended")
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_ERROR, envState.Status)
		assert.Equal(t,
			"bad thing happened: I just want not to work",
			envState.Errors[0].Text)
	}

	// Let time for postmortem service to do its job
	time.Sleep(250 * time.Millisecond)

	pmp, err := ft.Client.PMProcessQuery(
		context.Background(), api.LogLevel_WARNING, true, []api.UUID{processID}, true)

	if err != nil {
		t.Fatal(err)
	}

	if len(pmp) != 1 {
		t.Fatalf("Should get 1 and only 1 postmortem process. Got %v", pmp)
	}

	assert.Equal(t, processID, pmp[0].Process.ID)
	assert.Equal(t, api.PMProcess_NEW, pmp[0].Status)
	assert.Equal(t, api.LogLevel_ERROR, pmp[0].Level)

	if err := ft.Client.PMProcessSetStatus(processID, api.PMProcess_OPENED, "There"); err != nil {
		t.Fatal(err)
	}

	pmp, err = ft.Client.PMProcessQuery(
		context.Background(), api.LogLevel_WARNING, true, []api.UUID{processID}, true)

	if err != nil {
		t.Fatal(err)
	}

	if len(pmp) != 1 {
		t.Fatalf("Should get 1 and only 1 postmortem process. Got %v", pmp)
	}

	assert.Equal(t, processID, pmp[0].Process.ID)
	assert.Equal(t, api.PMProcess_OPENED, pmp[0].Status)
	assert.Equal(t, api.LogLevel_ERROR, pmp[0].Level)
	assert.Equal(t, "There", pmp[0].Comment)

	if err := ft.Client.PMProcessSetStatus(processID, api.PMProcess_CLOSED, "All done"); err != nil {
		t.Fatal(err)
	}

	pmp, err = ft.Client.PMProcessQuery(
		context.Background(), api.LogLevel_WARNING, false, []api.UUID{processID}, true)

	if err != nil {
		t.Fatal(err)
	}

	if len(pmp) != 0 {
		t.Fatalf("Closed pmprocess should not be returned. Got %v", pmp)
	}

	pmp, err = ft.Client.PMProcessQuery(
		context.Background(), api.LogLevel_WARNING, true, []api.UUID{processID}, true)

	if err != nil {
		t.Fatal(err)
	}

	if len(pmp) != 1 {
		t.Fatalf("Should get 1 and only 1 postmortem process. Got %v", pmp)
	}

	assert.Equal(t, processID, pmp[0].Process.ID)
	assert.Equal(t, api.PMProcess_CLOSED, pmp[0].Status)
	assert.Equal(t, api.LogLevel_ERROR, pmp[0].Level)
	assert.Equal(t, "All done", pmp[0].Comment)

}

func TestEndToEndSingleEventWithWarning(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	xsub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			t.Logf("Here i am")
			r, err := apc.ReadEnvelope("default")
			if err != nil {
				ft.T.Error(err)
				return err
			}
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return err
			}
			hit <- e
			apc.Log.EnvWarningf(
				apc.Request.Context,
				&api.Envelope{ID: r.ID},
				"some warning",
			)
			return nil
		})
	assert.Nil(t, err)
	defer xsub.Unsubscribe()

	assert.Equal(t, "jojo", consumer.Name)

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))

	env := xbus.ForgeEnvelopeFromItems("evt1", []api.Item{[]byte("\"item1\"")})

	envStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	status, err := emitter.SendEnvelope(nil, api.DefaultOutput, env)
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	timeout := time.After(5 * time.Second)
	var processID api.UUID
	select {
	case <-timeout:
		t.Errorf("Process state not updated")
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_RUNNING, envState.Status)
		processID = envState.ProcessID
	}

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("Process state not updated to 'done'")
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_DONE, envState.Status)
	}

	// Let time for postmortem service to do its job
	time.Sleep(100 * time.Millisecond)

	pmp, err := ft.Client.PMProcessQuery(
		context.Background(), api.LogLevel_WARNING, true, []api.UUID{processID}, true)

	if err != nil {
		t.Fatal(err)
	}

	if len(pmp) != 1 {
		t.Fatalf("Should get 1 and only 1 postmortem process. Got %v", pmp)
	}

	assert.Equal(t, processID, pmp[0].Process.ID)
	assert.Equal(t, api.PMProcess_NEW, pmp[0].Status)
	assert.Equal(t, api.LogLevel_WARNING, pmp[0].Level)
}
