// Copyright 2016 Orus.io. All rights reserved.

package fullstack

import (
	"fmt"
	"strconv"
	"sync"
	"sync/atomic"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

/* This test has no meaning with the new way envelopes are retrieved by the
actor.
However, once we have a back-pressure on envelope reading, we can recycle
this test to make sure slow or very slow actors are ok.

func TestSlowConsumer(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})

	ft.InstallGraph(`
name: forward
version: 0.1
status: active
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: target
  type: consumer
  roles: [anyconsumer]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> target
`)

	var (
		itemCount     int
		fragmentCount int
	)

	csub, err := consumer.RawSubscribe(
		func(input string, envelope api.Envelope) (api.EnvelopeAck, error) {
			time.Sleep(50 * time.Millisecond)
			ack := api.EnvelopeAck{
				ID:     envelope.ID,
				Status: api.EnvelopeAck_ACCEPTED,
			}
			fragmentCount++
			itemCount += len(envelope.Events[0].Items)
			t.Log("Consumer received envelope -", itemCount, "items so far")
			if itemCount == 3000 {
				go consumer.ProcessingSuccess(envelope.Context, envelope.ID)
			}
			return ack, nil
		}, true)
	if err != nil {
		t.Fatal(err)
	}
	defer csub.Unsubscribe()

	pStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	var item = make([]byte, 1024)

	go func() {
		output := emitter.OpenOutput(api.ProcessingContext{}, "default", "evt1")
		for i := 0; i != 3000; i++ {
			output.AddItems("evt1", api.Item(item))
		}
		assert.Nil(t, output.Close())
		t.Log("Sent all the items")
	}()

	timeout := time.After(20 * time.Second)
	var processID api.UUID

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_RUNNING, state.Status)
		processID = state.ProcessList[0].ID
	case <-timeout:
		ft.Logger.Errorf("Process never started")
		t.FailNow()
	}

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_DONE, state.Status)
		assert.Empty(t, state.ProcessList[0].Errors)
		pState := ft.Server.ProcessStorage().GetState(processID)
		t.Log(pState.Fmt())
	case <-timeout:
		ft.Logger.Errorf("Process never ended. Current state:")
		pState := ft.Server.ProcessStorage().GetState(processID)
		t.Log(pState.Fmt())
	}
}
*/

func TestStreamedEnvelope(t *testing.T) {
	const emissionLoops = 20000
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	cancel := make(chan struct{})

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})

	_, worker := ft.CreateWorker("fw", []string{"forward"})
	wsub, err := worker.Subscribe(func(apc *xbus.ActorProcessingContext) error {
		worker.Log.Noticef("Receiving envelope")
		r, err := apc.ReadEnvelope(api.DefaultInput)
		if err != nil {
			t.Error(err)
			return err
		}
		output := apc.OpenOutput("default", "evt1", "evt2")
		c1 := r.GetItemChan("evt1")
		c2 := r.GetItemChan("evt2")
		wg := sync.WaitGroup{}
		wg.Add(2)
		go func() {
			defer wg.Done()
			for {
				select {
				case <-cancel:
					return
				case item := <-c1:
					if item == nil {
						return
					}
					if err := output.AddItems("evt1", item); err != nil {
						t.Error(err)
						return
					}
				}
			}
		}()
		go func() {
			defer wg.Done()
			for {
				select {
				case <-cancel:
					return
				case item := <-c2:
					if item == nil {
						return
					}
					if err := output.AddItems("evt2", item); err != nil {
						t.Error(err)
						return
					}
				}
			}
		}()
		wg.Wait()
		worker.Log.Noticef("Done receiving")
		return output.Close()
	})
	defer wsub.Unsubscribe()

	var (
		receivedEvt1Count int64
		receivedEvt2Count int64
	)

	csub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			r, err := apc.ReadEnvelope(api.DefaultInput)
			if err != nil {
				t.Error(err)
				return err
			}
			wg := sync.WaitGroup{}
			wg.Add(2)
			c1 := r.GetItemChan("evt1")
			c2 := r.GetItemChan("evt2")
			go func() {
				defer wg.Done()
				for range c1 {
					atomic.AddInt64(&receivedEvt1Count, 1)
				}
			}()
			go func() {
				defer wg.Done()
				for range c2 {
					atomic.AddInt64(&receivedEvt2Count, 1)
				}
			}()
			select {
			case <-r.Done():
				wg.Wait()
			case <-cancel:
				return fmt.Errorf("Canceled")
			}
			ft.Logger.Noticef("consumer: done receiving")
			if r.LastError() != nil {
				ft.Logger.Noticef("consumer: return with err=%s", r.LastError())
				return r.LastError()
			}
			if atomic.LoadInt64(&receivedEvt1Count) != emissionLoops ||
				atomic.LoadInt64(&receivedEvt2Count) != 3*emissionLoops {
				err := fmt.Errorf(
					"Expected %d evt1, %d evt2 - Received: %d evt1, %d evt2",
					emissionLoops, 3*emissionLoops,
					receivedEvt1Count, receivedEvt2Count)
				return err
			}
			return nil
		})
	if err != nil {
		t.Fatal(err)
	}
	defer csub.Unsubscribe()

	ft.InstallGraph(`
name: forward
version: 0.1
status: active
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1,evt2]
  outputs: [default]
- id: relay
  type: worker
  actors: [fw]
  inputs: [default]
  outputs: [default]
- id: target
  type: consumer
  roles: [anyconsumer]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> relay
- relay.default -> target.default
`)

	pStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	go func() {
		output := emitter.OpenOutput(nil, "default", "evt1", "evt2")
		for i := 0; i != emissionLoops; i++ {
			output.AddItems("evt1", api.Item("evt1-item-"+strconv.Itoa(i)))
			output.AddItems("evt2",
				api.Item("evt2-0-item-"+strconv.Itoa(i)),
				api.Item("evt2-1-item-"+strconv.Itoa(i)),
				api.Item("evt2-2-item-"+strconv.Itoa(i)))
		}
		assert.Nil(t, output.Close())
	}()

	timeout := time.After(120 * time.Second)
	var processID api.UUID

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_RUNNING, state.Status)
		processID = state.ProcessID
	case <-timeout:
		ft.Logger.Errorf("Process never started")
		t.FailNow()
	}

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_DONE.String(), state.Status.String())
		assert.Empty(t, state.Errors)
		pState := ft.Server.ProcessStorage().GetState(processID)
		t.Log(pState.Fmt())
	case <-timeout:
		close(cancel)
		ft.Logger.Errorf("Process never ended. Current state:")
		pState := ft.Server.ProcessStorage().GetState(processID)
		t.Log(pState.Fmt())
	}
}
