package http

import (
	"encoding/json"
	"fmt"
	"net/http"
	"time"

	"github.com/gorilla/mux"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

// Beta11Envelope is a envelope structure from the beta11 version
type Beta11Envelope struct {
	api.Envelope
	Context *api.ProcessingContext `json:"context"`
}

// Beta11OutputHandler is the handler for the deprecated '{actor}/output/{output}' route
func (g *Gateway) Beta11OutputHandler(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	outputName := vars["output"]

	actor, err := g.getActor(r)
	if err != nil {
		w.WriteHeader(http.StatusNotFound)
		fmt.Fprintf(w, "Error: %s", err)
		return
	}
	defer actor.GetClient().Close()

	var envelope Beta11Envelope

	contentType := r.Header.Get("Content-Type")

	if contentType == "application/json" {
		decoder := json.NewDecoder(r.Body)
		if err := decoder.Decode(&envelope); err != nil {
			w.WriteHeader(http.StatusBadRequest)
			fmt.Fprintf(w, "Error decoding body: %s", err)
			return
		}
	} else {
		w.WriteHeader(http.StatusBadRequest)
		w.Write([]byte("Invalid content-type"))
	}

	var eventTypes []string
	var eventIDs []api.UUID
	for _, event := range envelope.Events {
		eventTypes = append(eventTypes, event.Type)
		eventIDs = append(eventIDs, event.ID)
	}
	sender := xbus.NewEnvelopeSenderWithIDs(
		envelope.ID,
		eventTypes,
		eventIDs,
		api.EnvelopeMaxSize,
		func(fragment *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
			return actor.SendEnvelope(envelope.Context, outputName, *fragment)
		},
	)
	for _, event := range envelope.Events {
		if err := sender.AddItems(event.Type, event.Items...); err != nil {
			w.WriteHeader(http.StatusInternalServerError)
			fmt.Fprintf(w, "Internal Error adding items: %s", err)
			return
		}
	}
	if err := sender.Close(); err != nil {
		w.WriteHeader(http.StatusInternalServerError)
		fmt.Fprintf(w, "Internal Error '%s'", err)
		return
	}

	w.WriteHeader(http.StatusOK)
	w.Write([]byte("ACK"))
}

// Beta11InputHandler is the handler for the deprecated '{actor}/input' route
func (g *Gateway) Beta11InputHandler(w http.ResponseWriter, r *http.Request) {
	actor, err := g.getActor(r)
	if err != nil {
		w.WriteHeader(http.StatusNotFound)
		fmt.Fprintf(w, "{\"error\": \"%s\"}", err)
		return
	}
	defer func() {
		// Without a delay, the last messages sent to the server are lost
		// FIXME this should not be necessary. Dig into nats to understand what
		// is going on. Before that, upgrade go-nats.
		time.AfterFunc(time.Second, actor.GetClient().Close)
	}()

	itemEncodings, err := api.ParseItemEncodingList(r.Header.Get("Accept-Item-Encoding"))
	if err != nil {
		w.WriteHeader(http.StatusBadRequest)
		fmt.Fprintf(w, "Invalid 'Accept-Item-Encoding': %s", err)
		return
	}
	if len(itemEncodings) == 0 {
		itemEncodings = []api.ItemEncoding{api.ItemBinary}
	}

	var (
		sub        *xbus.Subscription
		processing = make(chan struct{})
	)
	sub, err = actor.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			// Signal the http handler that we are processing something
			close(processing)

			apc.ActorLeaving()

			w.Header().Set("Content-Type", "application/json")

			b, err := api.JSONMarshal(
				itemEncodings, apc.Request.Inputs[0].Envelope,
			)

			if err != nil {
				w.WriteHeader(http.StatusInternalServerError)
				return err
			}

			m := make(map[string]interface{})

			if err := json.Unmarshal(b, &m); err != nil {
				w.WriteHeader(http.StatusInternalServerError)
				return err
			}

			m["context"] = apc.Request.Context

			m = map[string]interface{}{
				"input":    apc.Request.Inputs[0].Name,
				"envelope": m,
			}

			b, err = json.Marshal(m)

			if err != nil {
				w.WriteHeader(http.StatusInternalServerError)
				return err
			}

			w.WriteHeader(http.StatusOK)
			w.Write(b)

			// TODO If needed, send all the envelope remainings

			apc.Detach()
			return nil
		},
	)

	select {
	case <-r.Context().Done():
		if err := sub.Unsubscribe(); err != nil {
			g.Log.Errorf("Error unsubscribing the handler: %s", err)
		}
		return
	case <-processing:
		// The processing is ongoing, the sub.Unsubscribe will wait for its
		// completion
		if err := sub.Unsubscribe(); err != nil {
			g.Log.Errorf("Error unsubscribing the handler: %s", err)
		}

	case <-time.After(30 * time.Second):
		// Unsubscribe, then check if a processing was done meanwhile
		if err := sub.Unsubscribe(); err != nil {
			g.Log.Errorf("Error unsubscribing the handler: %s", err)
		}

		select {
		case <-processing:
			// Yes, so we have nothing to write back
		default:
			w.WriteHeader(http.StatusNoContent)
		}
	}
}
