package license

/* This package provides some binary licensing tools.
By default, calling the functions of this package has no effect.

If license, licenseWarning and/or licenseExpiration are defined at link time,
however, the functions will start to do something.

In any case, the program feature are _not_ modified in any way aside from
the addition of a 'license' subcommand, and the display of license information
on stdout and the logs in case if needed.
*/

import (
	"fmt"
	"strings"
	"time"

	"bitbucket.org/orus-io/go-xbus"
	"github.com/spf13/cobra"
)

var (
	license           string
	licenseWarning    string
	licenseExpiration string

	licenseExpirationDate time.Time
)

// DisplayLicenseOnStart display the license at program start
// returns true if the binary has a usage license
func DisplayLicenseOnStart() bool {
	if license != "" {
		if !licenseExpirationDate.IsZero() {
			if time.Now().After(licenseExpirationDate) {
				fmt.Println(license)
				fmt.Println(licenseWarning)
				fmt.Println("Execution will continue normally in 5 seconds")
				time.Sleep(5 * time.Second)
			}
		} else {
			fmt.Println(license)
		}
		return true
	}
	return false
}

// SendExpirationError send an error to the logger if the license has expired
func SendExpirationError(logger xbus.Logger) {
	logger.Errorf("%s", licenseWarning)
}

// SetupLicenseWatcher spawn a routine that will emit an error every 15 minutes
// if the license has expired
func SetupLicenseWatcher(logger xbus.Logger) {
	if licenseExpirationDate.IsZero() {
		return
	}
	go func() {
		for licenseExpirationDate.After(time.Now()) {
			<-time.After(time.Now().Sub(licenseExpirationDate))
		}
		SendExpirationError(logger)
		t := time.Tick(15 * time.Minute)
		for range t {
			SendExpirationError(logger)
		}
	}()
}

var licenseCmd = cobra.Command{
	Use:   "license",
	Short: "Show this binary usage license",
	Run: func(cmd *cobra.Command, args []string) {
		if licenseExpirationDate.IsZero() ||
			licenseExpirationDate.Before(time.Now()) {
			// The license should have been displayed at startup already
			return
		}
		fmt.Println(license)
	},
}

// SetupCmd adds a 'license' command if applicable
func SetupCmd(rootCmd *cobra.Command) {
	if license != "" {
		rootCmd.AddCommand(&licenseCmd)
	}
}

func init() {
	if license != "" {
		license = strings.Replace(license, "\\n", "\n", -1)
	}
	if licenseWarning != "" {
		licenseWarning = strings.Replace(licenseWarning, "\\n", "\n", -1)
	}
	if licenseExpiration != "" {
		t, err := time.Parse("2006-01-02", licenseExpiration)
		if err != nil {
			panic(err)
		}
		licenseExpirationDate = t
	}
}
