package testutils

import (
	"context"
	"fmt"
	"sync"
	"testing"
	"time"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

const (
	// DefaultNodeID used in the ProcessingContext
	DefaultNodeID = "node"

	// DefaultActorName of the test actor
	DefaultActorName = "actor"
)

var (
	// DefaultProcessID used in the ProcessingContext
	DefaultProcessID = api.MustUUIDFromString("af75efb0-1f0d-11e9-9513-2329a2a94eaf")
	// DefaultActorID of the test actor
	DefaultActorID = api.MustUUIDFromString("7fe4f2b8-1f0e-11e9-8661-ebbdab78f251")
)

// ActorServiceTester helps testing a ActorProcessService service
type ActorServiceTester struct {
	serviceName string
	factory     xbus.NewActorServiceFunc
	Apc         xbus.ActorProcessingContext
	sendChannel chan api.ActorProcessingState

	Outputs map[string]*xbus.EnvelopeReceiver
}

// NewActorServiceTester creates a ActorServiceTester
func NewActorServiceTester(
	t *testing.T,
	serviceName string,
	factory func(actor *xbus.Actor) xbus.ActorService,
) *ActorServiceTester {
	var tester = ActorServiceTester{
		serviceName: serviceName,
		factory:     factory,
		Apc: xbus.ActorProcessingContext{
			Request: &api.ActorProcessRequest{
				Context: &api.ProcessingContext{
					ProcessID: DefaultProcessID,
					NodeID:    DefaultNodeID,
				},
			},
			Actor: &xbus.Actor{
				ID:   DefaultActorID,
				Name: DefaultActorName,
			},
		},
		Outputs: make(map[string]*xbus.EnvelopeReceiver),
	}
	xbus.ApcTestSetSend(
		&tester.Apc,
		func(s api.ActorProcessingState) { tester.sendChannel <- s },
	)
	xbus.ApcTestSetRetrieveEnvelopeFrom(
		&tester.Apc,
		func(
			ctx context.Context, clientKind, clientID string,
			envelope api.Envelope, position api.EnvelopePosition,
		) *xbus.EnvelopeReceiver {
			receiver, err := xbus.NewEnvelopeReceiver("", envelope)
			if err != nil {
				panic(err)
			}
			// TODO connect FeedInput to receiver.AddFragment
			if receiver.Status() != api.EnvelopeAck_ACCEPTED {
				panic("Envelope must be complete")
			}
			return receiver
		},
	)
	xbus.ApcTestSetOpenOutput(
		&tester.Apc,
		func(
			context *api.ProcessingContext, output string, eventTypes ...string,
		) *xbus.EnvelopeSender {
			receiver := tester.getOutputReceiver(output)
			sender := xbus.NewEnvelopeSender(
				eventTypes, 1024*512,
				func(e *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
					receiver.AddFragment(*e)
					return receiver.Status(), receiver.LastError()
				})
			return sender
		},
	)
	xbus.ApcTestSetCloseOutput(
		&tester.Apc,
		func(
			context *api.ProcessingContext, output string,
		) (api.EnvelopeAck, error) {
			return api.EnvelopeAck{
				Status: api.EnvelopeAck_ACCEPTED,
			}, tester.getOutputReceiver(output).Finalize()
		},
	)
	xbus.ApcTestSetSendEnvelope(
		&tester.Apc,
		func(
			ctx *api.ProcessingContext, output string, env api.Envelope,
		) (api.EnvelopeAck_ReceptionStatus, error) {
			receiver := tester.getOutputReceiver(output)
			receiver.AddFragment(env)
			return receiver.Status(), receiver.LastError()
		},
	)
	xbus.ApcTestSetProcessingEnd(
		&tester.Apc,
		func(
			context *api.ProcessingContext,
			status api.ActorProcessingState_Status,
			messages []api.LogMessage,
		) error {
			tester.sendChannel <- api.ActorProcessingState{
				Context:  *context,
				Status:   status,
				Messages: messages,
			}
			return nil
		},
	)
	return &tester
}

// SetSettings sets the service settings in the Actor definition
func (t *ActorServiceTester) SetSettings(settings map[string]interface{}) {
	t.Apc.Actor.Options.Service.Settings = settings
}

// SetInputEnvelope associate an initial envelope fragment to an input
// If needed, further fragments can be fed during the RunProcess() with FeedInput
func (t *ActorServiceTester) SetInputEnvelope(name string, fragment *api.Envelope) {
	t.Apc.Request.Inputs = append(t.Apc.Request.Inputs, &api.ActorProcessRequest_Input{
		Name:     name,
		Envelope: fragment,
		Position: &api.EnvelopePosition{},
	})
}

// CloseInput marks an input as closed
func (t *ActorServiceTester) CloseInput(name string) {
	t.Apc.Request.Inputs = append(t.Apc.Request.Inputs, &api.ActorProcessRequest_Input{
		Name:  name,
		Close: true,
	})
}

// FeedInput adds a fragment to an input stream. Must be called _during_ Run to
// avoid being locked
func (t *ActorServiceTester) FeedInput(name string, fragment api.Envelope) {
	panic("Not yet implemented")
}

// RunProcess runs Process() with the inputs if provided and store the outputs for further inspection
func (t *ActorServiceTester) RunProcess() (*api.ActorProcessingState, error) {
	var lastState *api.ActorProcessingState

	var baseService xbus.ActorService
	if t.factory != nil {
		baseService = t.factory(t.Apc.Actor)
	} else {
		baseService = xbus.MakeActorService(t.serviceName, t.Apc.Actor)
	}
	if baseService == nil {
		return nil, fmt.Errorf("Factory returned nil")
	}
	service, ok := baseService.(xbus.ActorProcessService)
	if !ok {
		return nil, fmt.Errorf("Factory returned a non ActorProcessService")
	}
	t.sendChannel = make(chan api.ActorProcessingState, 1)

	wg := sync.WaitGroup{}
	wg.Add(1)
	go func() {
		for state := range t.sendChannel {
			lastState = &state
		}
		wg.Done()
	}()

	err := service.Process(&t.Apc)

	close(t.sendChannel)
	wg.Wait()

	if err != nil {
		if t.Apc.IsDone() {
			return lastState, fmt.Errorf("Process() returned an error but is marked as done")
		}
		if t.Apc.IsDetached() {
			return lastState, fmt.Errorf("Process() returned an error but is marked as detached")
		}
		if lastState == nil {
			lastState = &api.ActorProcessingState{
				Status: api.ActorProcessingState_ERROR,
				Messages: []api.LogMessage{
					{
						Time:  time.Now(),
						Level: api.LogLevel_ERROR,
						Text:  err.Error(),
					},
				},
			}
		} else {
			lastState.Status = api.ActorProcessingState_ERROR
		}
	} else if lastState == nil {
		if !t.Apc.IsDetached() {
			lastState = &api.ActorProcessingState{
				Status: api.ActorProcessingState_SUCCESS,
			}
		}
	}

	return lastState, nil
}

func (t *ActorServiceTester) getOutputReceiver(name string) *xbus.EnvelopeReceiver {
	receiver, ok := t.Outputs[name]
	if !ok {
		receiver = xbus.NewEnvelopeReceiverEmpty()
		t.Outputs[name] = receiver
	}
	return receiver
}

// GetOutputEnvelope returns the complete envelope send on a given output
// Run must be over
func (t *ActorServiceTester) GetOutputEnvelope(name string, timeout time.Duration) (*api.Envelope, error) {
	return t.getOutputReceiver(name).Complete(timeout)
}

// GetOutputReceiver returns an envelope receiver for envelope send on a given
// output
func (t *ActorServiceTester) GetOutputReceiver(name string) *xbus.EnvelopeReceiver {
	return t.getOutputReceiver(name)
}
